from Bio import Phylo
from iterpop import iterpop as ip
from lyncs_utils import keydefaultdict
from nanto import isanan, nantonone
import pandas as pd
import string
import typing


def _parse_ancestor_list(raw: str) -> typing.List[int]:
    assert not any(
        c in raw for c in string.whitespace
    ), "Whitespace separated ancestor list not supported."
    try:
        ancestor_list = eval(raw)
        if ancestor_list == [None]:
            return []
        else:
            assert None not in ancestor_list
            return ancestor_list
    except NameError:
        # if ancestor list contains a placeholder string like NONE,
        # it will trigger a NameError when eval'ed
        return []


def alife_dataframe_to_biopython_trees(
    df: pd.DataFrame,
    setattrs: typing.Optional[typing.Union[
        typing.Iterable[str],
        typing.Mapping[str, str],
    ]] = None,
    setup_branch_lengths: bool = False,
):
    """Open a phylogeny dataframe formatted to the artificial life community
    data format standards as zero or more biopython trees, depending on the
    number of clades with no common ancestor.

    The following columns will automatically be applied as attributes to
    generated Clade objects:
        * branch_length,
        * id,
        * name, and
        * origin_time.

    Parameters
    ----------
    df:
        Pandas DataFrame to convert.
    setattrs: optional
        Dataframe columns that should be attached as attributes to Clade
        objects within the trees. If a map is provided, values at columns in
        keys will be attached with the corresponding value as the attr name.
    setup_branch_lengths: bool, optional
        Should we try to set up branch lengths using the origin_time column?
        Will not override if branch_length is provided as a column of df.
    """

    df = df.copy()

    df['parsed_ancestor_list'] \
        = df['ancestor_list'].apply(_parse_ancestor_list)

    # maps id to node
    def setup_node(id: int) -> Phylo.BaseTree.Clade:
        res = Phylo.BaseTree.Clade()
        res.id = id
        return res
    nodes = keydefaultdict(setup_node)

    # maps id to origin time
    root_nodes = []

    for __, row in df.iterrows():
        node = nodes[row['id']]

        if setattrs is not None:
            for col_name in setattrs:
                attr_name: str
                try:
                    attr_name = setattrs[col_name]
                except TypeError:
                    attr_name = col_name
                assert attr_name not in (
                    'name',
                    'branch_length',
                    'origin_time',
                )
                assert not hasattr(node, attr_name)
                setattr(node, attr_name, row[col_name])

        node.origin_time = nantonone(row.get('origin_time', default=None))
        node.name = row.get('name', default=None)
        node.branch_length = nantonone(row.get('branch_length', None))

        if row['parsed_ancestor_list']:
            ancestor_id = ip.popsingleton(row['parsed_ancestor_list'])
            nodes[ancestor_id].clades.append(node)
        else:
            root_nodes.append(node)

    # set up branch lengths
    if setup_branch_lengths:
        for id, parent_node in nodes.items():
            for child_node in parent_node.clades:
                if child_node.branch_length is None and None not in (
                    getattr(child_node, 'origin_time', None),
                    getattr(parent_node, 'origin_time', None),
                ):
                    assert child_node.origin_time >= parent_node.origin_time
                    child_node.branch_length \
                        = child_node.origin_time - parent_node.origin_time

        for root_node in root_nodes:
            if (
                root_node.branch_length is None
                and getattr(root_node, 'origin_time', None) is not None
            ):
                assert not isanan(root_node.origin_time)
                root_node.branch_length = root_node.origin_time

    return([
        Phylo.BaseTree.Tree(root=root_node)
        for root_node in root_nodes
    ])



#!/usr/bin/env python

'''
`dendropy_tree_to_alife_dataframe` tests for
`alifedata-phyloinformatics-convert` package.
'''

import dendropy
from dendropy.calculate import treecompare
from iterpop import iterpop as ip
from os.path import dirname, realpath

import alifedata_phyloinformatics_convert as apc


def test_newick1():

    original_tree = dendropy.Tree.get(
        path=f'{dirname(realpath(__file__))}/assets/APG_Angiosperms.newick',
        schema='newick',
    )

    converted_df = apc.dendropy_tree_to_alife_dataframe(original_tree)
    reconverted_tree = apc.alife_dataframe_to_dendropy_tree(converted_df)
    reconverted_df = apc.dendropy_tree_to_alife_dataframe(reconverted_tree)

    assert len(converted_df.compare(reconverted_df)) == 0
    assert len(converted_df) == len(reconverted_df)
    assert len(converted_df) == len(original_tree.nodes())
    assert len(reconverted_df) == len(original_tree.nodes())
    assert len(reconverted_df) == len(reconverted_tree.nodes())

    assert len(reconverted_tree.nodes()) == len(original_tree.nodes())
    assert len(reconverted_tree.leaf_nodes()) \
        == len(original_tree.leaf_nodes())
    assert original_tree.as_ascii_plot() == reconverted_tree.as_ascii_plot()
    assert str(original_tree) == str(reconverted_tree)


def test_newick2():

    original_tree = dendropy.Tree.get(
        path=f'{dirname(realpath(__file__))}/assets/APG_Angiosperms.newick',
        schema='newick',
    )
    original_tree.resolve_polytomies()
    original_tree.suppress_unifurcations()
    original_tree.collapse_basal_bifurcation()

    converted_df = apc.dendropy_tree_to_alife_dataframe(original_tree)
    reconverted_tree = apc.alife_dataframe_to_dendropy_tree(converted_df)
    reconverted_df = apc.dendropy_tree_to_alife_dataframe(reconverted_tree)

    assert len(converted_df.compare(reconverted_df)) == 0
    assert len(converted_df) == len(reconverted_df)
    assert len(converted_df) == len(original_tree.nodes())
    assert len(reconverted_df) == len(original_tree.nodes())
    assert len(reconverted_df) == len(reconverted_tree.nodes())

    reconverted_tree.migrate_taxon_namespace(original_tree.taxon_namespace)

    assert original_tree.as_ascii_plot() == reconverted_tree.as_ascii_plot()
    assert str(original_tree) == str(reconverted_tree)
    assert treecompare.symmetric_difference(
        original_tree,
        reconverted_tree,
    ) == 0


def test_nexml():

    script_directory = dirname(realpath(__file__))
    original_tree = dendropy.Tree.get(
        path=f'{script_directory}/assets/pythonidae.annotated.nexml',
        schema='nexml',
    )
    converted_df = apc.dendropy_tree_to_alife_dataframe(original_tree)
    reconverted_tree = apc.alife_dataframe_to_dendropy_tree(
        converted_df,
    )

    # ensure taxons can be compared between trees
    reconverted_tree.migrate_taxon_namespace(original_tree.taxon_namespace)

    assert treecompare.symmetric_difference(
        original_tree,
        reconverted_tree,
    ) == 0


def test_exportattrs_iterable():

    script_directory = dirname(realpath(__file__))
    original_tree = dendropy.Tree.get(
        path=f'{script_directory}/assets/pythonidae.annotated.nexml',
        schema='nexml',
    )

    for node in original_tree.leaf_node_iter():
        node.fish = 'Salmon'

    for node in original_tree.preorder_internal_node_iter():
        node.fish = 'Tilapia'

    converted_df = apc.dendropy_tree_to_alife_dataframe(
        original_tree,
        exportattrs=['fish'],
    )

    assert 'fish' in converted_df
    assert sum(converted_df['fish'] == 'Salmon') \
        == len(original_tree.leaf_nodes())
    assert sum(converted_df['fish'] == 'Tilapia') \
        == len(original_tree.internal_nodes())


def test_exportattrs_mapping():

    script_directory = dirname(realpath(__file__))
    original_tree = dendropy.Tree.get(
        path=f'{script_directory}/assets/pythonidae.annotated.nexml',
        schema='nexml',
    )
    for i, node in enumerate(original_tree):
        if node.taxon is None:
            node.taxon = original_tree.taxon_namespace.new_taxon(str(i))

    for node in original_tree.leaf_node_iter():
        node.fish = 'Salmon'
        node.soup = None

    for node in original_tree.preorder_internal_node_iter():
        node.fish = 'Tilapia'
        node.soup = None

    converted_df = apc.dendropy_tree_to_alife_dataframe(
        original_tree,
        exportattrs={'fish': 'The Fish', 'soup': 'soup', 'taxon.label': 'name'},
    )

    assert 'The Fish' in converted_df
    assert sum(converted_df['The Fish'] == 'Salmon') \
        == len(original_tree.leaf_nodes())
    assert sum(converted_df['The Fish'] == 'Tilapia') \
        == len(original_tree.internal_nodes())

    assert 'soup' in converted_df
    assert ip.popsingleton(converted_df['soup'].unique()) is None

    assert 'name' in converted_df
    assert {*converted_df['name']} ==  {node.taxon.label for node in original_tree}
    assert len(converted_df['name'].unique()) > 1



from Bio import Phylo
from collections import defaultdict
from collections.abc import Mapping
import functools
from nanto import isanan
import opytional as opyt
import pandas as pd
import typing


# adapted from https://stackoverflow.com/a/31174427/17332200
def _rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return functools.reduce(_getattr, [obj] + attr.split('.'))


def biopython_tree_to_alife_dataframe(
    tree: Phylo.BaseTree,
    exportattrs: typing.Optional[typing.Union[
        typing.Iterable[str],
        typing.Mapping[str, str],
    ]] = None,
) -> pd.DataFrame:
    """Convert a biopython phylogenetic tree to a dataframe formatted to the
    artificial life communit data format standards.

    The following Clade object attributes will automatically be exported to
    dataframe columns, if available:
        * branch_length,
        * id,
        * name, and
        * origin_time.

    Parameters
    ----------
    tree:
        biopython tree to convert.
    exportattrs: optional
        Clade attrs that should be copied as columns into the generated
        dataframe. If a map is provided, attr values in keys will be inserted
        into the dataframe with the corresponding value as the column name.
    """

    # adapted from https://biopython.org/wiki/Phylo_cookbook
    def all_parents(tree):
        parents = defaultdict(lambda: None)
        for clade in tree.find_clades(order='level'):
            for child in clade:
                parents[child] = clade
        return parents
    parents = all_parents(tree)

    # set up clade origin times if any edge lengths set
    if any(clade.branch_length is not None for clade in tree.find_clades()):
        if not hasattr(tree.root, 'origin_time'):
            tree.root.origin_time = opyt.or_value(
                tree.root.branch_length,
                0
            )
        else:
            assert tree.root.origin_time is not None
        for clade in tree.find_clades(order='level'):
            parent = parents[clade]
            if parent is not None and not hasattr(clade, 'origin_time'):
                if None not in (
                    getattr(parent, 'origin_time', None),
                    clade.branch_length,
                ) and not isanan(parent.origin_time):
                    clade.origin_time \
                        = parent.origin_time + clade.branch_length

    # attach ids to clades, if needed
    for fallback_id, clade in enumerate(tree.find_clades(order='preorder')):
        if not hasattr(clade, 'id'):
            clade.id = fallback_id
        else:
            assert isinstance(clade.id, int)

    assert not any((
        attr in (
            exportattrs.values()
            if isinstance(exportattrs, Mapping)
            else opyt.or_value(exportattrs, [])
        )
        for attr in ('origin_time', 'id', 'branch_length', 'name')
    ))
    return pd.DataFrame.from_records([
        {
            **{
                'id': clade.id,
                'ancestor_list': str(
                    [opyt.apply_if(parents[clade], lambda x: x.id)]
                ),
                'origin_time': getattr(clade, 'origin_time', None),
                'branch_length': clade.branch_length,
                'name': clade.name,
            },
            **{
                exportattrs[attr_name]
                if isinstance(exportattrs, Mapping)
                else attr_name: _rgetattr(clade, attr_name)
                for attr_name in opyt.or_value(exportattrs, [])
            },
        }
        for clade in tree.find_clades(order='preorder')
    ])



#!/usr/bin/env python

'''
`dendropy_tree_to_alife_dataframe` tests for
`alifedata-phyloinformatics-convert` package.
'''

from Bio import Phylo
import dendropy
from iterpop import iterpop as ip
from os.path import dirname, realpath

import alifedata_phyloinformatics_convert as apc


def test_newick():

    dendropy_tree = dendropy.Tree.get(
        path=f'{dirname(realpath(__file__))}/assets/APG_Angiosperms.newick',
        schema='newick',
    )
    dendropy_df = apc.dendropy_tree_to_alife_dataframe(dendropy_tree)

    biopython_tree = Phylo.read(
        f'{dirname(realpath(__file__))}/assets/APG_Angiosperms.newick',
        'newick',
    )
    biopython_df = apc.biopython_tree_to_alife_dataframe(biopython_tree)

    assert len(dendropy_df) == len(biopython_df)

    assert len(dendropy_df['id'].compare(biopython_df['id'])) == 0
    assert len(dendropy_df['ancestor_list'].compare(
        biopython_df['ancestor_list']),
    ) == 0
    assert len(dendropy_df['origin_time'].compare(
        biopython_df['origin_time']),
    ) == 0
    assert len(dendropy_df['edge_length'].compare(
        biopython_df['branch_length']),
    ) == 0


def test_nexml():

    script_directory = dirname(realpath(__file__))
    dendropy_tree = dendropy.Tree.get(
        path=f'{script_directory}/assets/pythonidae.annotated.nexml',
        schema='nexml',
    )
    dendropy_df = apc.dendropy_tree_to_alife_dataframe(dendropy_tree)

    biopython_tree = Phylo.read(
        f'{script_directory}/assets/pythonidae.annotated.nexml',
        'nexml',
    )
    biopython_df = apc.biopython_tree_to_alife_dataframe(biopython_tree)
    assert len(dendropy_df) == len(biopython_df)

    assert len(dendropy_df) == len(biopython_df)


def test_exportattrs_iterable():

    original_tree = Phylo.read(
        f'{dirname(realpath(__file__))}/assets/APG_Angiosperms.newick',
        'newick',
    )

    for node in original_tree.get_terminals():
        node.fish = 'Salmon'

    for node in original_tree.get_nonterminals():
        node.fish = 'Tilapia'

    converted_df = apc.biopython_tree_to_alife_dataframe(
        original_tree,
        exportattrs=['fish'],
    )

    assert 'fish' in converted_df
    assert sum(converted_df['fish'] == 'Salmon') \
        == len(original_tree.get_terminals())
    assert sum(converted_df['fish'] == 'Tilapia') \
        == len(original_tree.get_nonterminals())


def test_exportattrs_mapping():

    original_tree = Phylo.read(
        f'{dirname(realpath(__file__))}/assets/APG_Angiosperms.newick',
        'newick',
    )

    for clade in original_tree.get_terminals():
        clade.fish = 'Salmon'
        clade.soup = None

    for clade in original_tree.get_nonterminals():
        clade.fish = 'Tilapia'
        clade.soup = None

    converted_df = apc.biopython_tree_to_alife_dataframe(
        original_tree,
        exportattrs={'fish': 'The Fish', 'soup': 'soup', 'name': 'taxon_label'},
    )

    assert 'The Fish' in converted_df
    assert sum(converted_df['The Fish'] == 'Salmon') \
        == len(original_tree.get_terminals())
    assert sum(converted_df['The Fish'] == 'Tilapia') \
        == len(original_tree.get_nonterminals())

    assert 'soup' in converted_df
    assert ip.popsingleton(converted_df['soup'].unique()) is None

    assert 'taxon_label' in converted_df
    assert (
        {*converted_df['taxon_label']}
        == {node.name for node in original_tree.find_clades()}
    )
    assert len(converted_df['taxon_label'].unique()) > 1



from collections import defaultdict
import dendropy
from nanto import nantonone
import numpy as np
import typing


def scipy_linkage_matrix_to_dendropy_tree(
    matrix: np.array,
    leaf_labels: typing.Optional[typing.List] = None,
    label_leaf_nodes: bool = True,
    label_leaf_taxa: bool = True,
) -> dendropy.Tree:
    """
    Parameters
    ----------
    leaf_labels: list, optional
        Labels for leaves of tree. If provided, must exactly equal the number
        of leaf nodes (i.e., num matrix rows + 1).
    label_leaf_nodes: bool, default True
        Apply leaf labels directly to leaf nodes.
    label_leaf_taxa: bool, default True
        Apply leaf labels to leaf node taxa.
    """

    # scipy linkage format
    # http://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html#scipy.cluster.hierarchy.linkage
    num_rows = len(matrix)

    # cluster id -> node
    nodes = defaultdict(dendropy.Node)

    if leaf_labels is not None:
        assert len(leaf_labels) == num_rows + 1
        for cluster_id, label in enumerate(leaf_labels):
            if label_leaf_taxa:
                nodes[cluster_id].taxon = dendropy.Taxon(label=label)
            if label_leaf_nodes:
                nodes[cluster_id].label = label

    for row_idx, row in enumerate(matrix):
        parent_cluster = row_idx + num_rows + 1
        joined_cluster1, joined_cluster2, cluster_distance, cluster_size = row

        nodes[parent_cluster].cluster_id = parent_cluster
        for child_cluster in joined_cluster1, joined_cluster2:
            nodes[parent_cluster].add_child(nodes[child_cluster])
            nodes[child_cluster].cluster_id = child_cluster
            nodes[child_cluster].edge_length = nantonone(cluster_distance / 2)

    return dendropy.Tree(seed_node=nodes[parent_cluster])



#!/usr/bin/env python

'''
`dendropy_tree_to_scipy_linkage_matrix` tests for
`alifedata-phyloinformatics-convert` package.
'''

import numpy as np

import alifedata_phyloinformatics_convert as apc


def test():

    # from https://stackoverflow.com/a/40983611/17332200
    linkage_matrix = np.array([
        [7.0, 9.0, 0.3, 2.0],
        [4.0, 6.0, 0.5, 2.0],
        [5.0, 12.0, 0.5, 3.0],
        [2.0, 13.0, 0.53851648, 4.0],
        [3.0, 14.0, 0.58309519, 5.0],
        [1.0, 15.0, 0.64031242, 6.0],
        [10.0, 11.0, 0.72801099, 3.0],
        [8.0, 17.0, 1.2083046, 4.0],
        [0.0, 16.0, 1.5132746, 7.0],
        [18.0, 19.0, 1.92353841, 11.0],
    ])

    tree = apc.scipy_linkage_matrix_to_dendropy_tree(linkage_matrix)
    reconstructed_linkage_matrix \
        = apc.dendropy_tree_to_scipy_linkage_matrix(tree)

    assert set(map(tuple, linkage_matrix)) \
        == set(map(tuple, reconstructed_linkage_matrix))


def test_no_cluster_ids():
    # from https://stackoverflow.com/a/40983611/17332200
    linkage_matrix = np.array([
        [7.0, 9.0, 0.3, 2.0],
        [4.0, 6.0, 0.5, 2.0],
        [5.0, 12.0, 0.5, 3.0],
        [2.0, 13.0, 0.53851648, 4.0],
        [3.0, 14.0, 0.58309519, 5.0],
        [1.0, 15.0, 0.64031242, 6.0],
        [10.0, 11.0, 0.72801099, 3.0],
        [8.0, 17.0, 1.2083046, 4.0],
        [0.0, 16.0, 1.5132746, 7.0],
        [18.0, 19.0, 1.92353841, 11.0],
    ])

    tree = apc.scipy_linkage_matrix_to_dendropy_tree(linkage_matrix)
    for node in tree:
        del node.cluster_id
    reconstructed_linkage_matrix \
        = apc.dendropy_tree_to_scipy_linkage_matrix(tree)

    assert [*linkage_matrix[:, 2]] == [*reconstructed_linkage_matrix[:, 2]]
    assert [*linkage_matrix[:, 3]] == [*reconstructed_linkage_matrix[:, 3]]



#!/usr/bin/env python

'''
`alife_dataframe_to_dendropy_tree` tests for
`alifedata-phyloinformatics-convert` package.
'''

from click.testing import CliRunner
import filecmp
from os.path import dirname, realpath
import tempfile

from alifedata_phyloinformatics_convert import cli


def test_toalifedata_nexml_csv():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.toalifedata,
            f'--input-file {scriptdir}/assets/pythonidae.annotated.nexml '
            '--input-schema nexml '
            f'--output-file {tempdir}/pythonidae.annotated.csv '
            '--output-format csv '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_toalifedata/pythonidae.annotated.csv',
            f'{tempdir}/pythonidae.annotated.csv',
        )


def test_toalifedata_nexml_json():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.toalifedata,
            f'--input-file {scriptdir}/assets/pythonidae.annotated.nexml '
            '--input-schema nexml '
            f'--output-file {tempdir}/pythonidae.annotated.json '
            '--output-format json '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_toalifedata/pythonidae.annotated.json',
            f'{tempdir}/pythonidae.annotated.json',
        )


def test_toalifedata_newick_csv():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.toalifedata,
            f'--input-file {scriptdir}/assets/APG_Angiosperms.newick '
            '--input-schema newick '
            f'--output-file {tempdir}/APG_Angiosperms.csv '
            '--output-format csv '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_toalifedata/APG_Angiosperms.csv',
            f'{tempdir}/APG_Angiosperms.csv',
        )


def test_toalifedata_newick_json():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.toalifedata,
            f'--input-file {scriptdir}/assets/APG_Angiosperms.newick '
            '--input-schema newick '
            f'--output-file {tempdir}/APG_Angiosperms.json '
            '--output-format json '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_toalifedata/APG_Angiosperms.json',
            f'{tempdir}/APG_Angiosperms.json',
        )


def test_fromalifedata_newick():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.fromalifedata,
            f'--input-file {scriptdir}/assets/alifedata.csv '
            '--input-format csv '
            f'--output-file {tempdir}/alifedata.newick '
            '--output-schema newick '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_fromalifedata/alifedata.newick',
            f'{tempdir}/alifedata.newick',
        )


def test_fromalifedata_minimal_newick():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.fromalifedata,
            f'--input-file {scriptdir}/assets/alifedata_minimal.csv '
            '--input-format csv '
            f'--output-file {tempdir}/alifedata_minimal.newick '
            '--output-schema newick '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_fromalifedata/alifedata_minimal.newick',
            f'{tempdir}/alifedata_minimal.newick',
        )


def test_fromalifedata_nexml():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.fromalifedata,
            f'--input-file {scriptdir}/assets/alifedata.csv '
            '--input-format csv '
            f'--output-file {tempdir}/alifedata.nexml '
            '--output-schema nexml '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_fromalifedata/alifedata.nexml',
            f'{tempdir}/alifedata.nexml',
        )


def test_fromalifedata_minimal_nexml():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.fromalifedata,
            f'--input-file {scriptdir}/assets/alifedata_minimal.csv '
            '--input-format csv '
            f'--output-file {tempdir}/alifedata_minimal.nexml '
            '--output-schema nexml '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_fromalifedata/alifedata_minimal.nexml',
            f'{tempdir}/alifedata_minimal.nexml',
        )


def test_fromalifedata_nexus():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.fromalifedata,
            f'--input-file {scriptdir}/assets/alifedata.csv '
            '--input-format csv '
            f'--output-file {tempdir}/alifedata.nexus '
            '--output-schema nexus '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_fromalifedata/alifedata.nexus',
            f'{tempdir}/alifedata.nexus',
        )


def test_fromalifedata_minimal_nexus():
    runner = CliRunner()
    scriptdir = dirname(realpath(__file__))
    with tempfile.TemporaryDirectory() as tempdir:
        result = runner.invoke(
            cli.fromalifedata,
            f'--input-file {scriptdir}/assets/alifedata_minimal.csv '
            '--input-format csv '
            f'--output-file {tempdir}/alifedata_minimal.nexus '
            '--output-schema nexus '
        )
        assert result.exit_code == 0
        assert filecmp.cmp(
            f'{scriptdir}/converted_fromalifedata/alifedata_minimal.nexus',
            f'{tempdir}/alifedata_minimal.nexus',
        )



from collections.abc import Mapping
import dendropy
import functools
from nanto import isanan
import opytional as opyt
import pandas as pd
import typing


# adapted from https://stackoverflow.com/a/31174427/17332200
def _rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return functools.reduce(_getattr, [obj] + attr.split('.'))


def dendropy_tree_to_alife_dataframe(
    tree: dendropy.Tree,
    exportattrs: typing.Optional[typing.Union[
        typing.Iterable[str],
        typing.Mapping[str, str],
    ]] = None,
) -> pd.DataFrame:
    """Convert a dendropy phylogenetic tree to a dataframe formatted to the
    artificial life communit data format standards.

    The following Node object attributes will automatically be exported to
    dataframe columns, if available:
        * edge_length,
        * id,
        * label,
        * origin_time, and
        * taxon.label.

    Parameters
    ----------
    tree:
        dendropy tree to convert.
    exportattrs: optional
        Node attrs that should be copied as columns into the generated
        dataframe. If a map is provided, attr values in keys will be inserted
        into the dataframe with the corresponding value as the column name.
    """

    # set up node origin times if any edge lengths set
    if any(node.edge_length is not None for node in tree):
        if not hasattr(tree.seed_node, 'origin_time'):
            tree.seed_node.origin_time = opyt.or_value(
                tree.seed_node.edge_length,
                0
            )
        else:
            assert tree.seed_node.origin_time is not None
        for node in tree:
            parent = node.parent_node
            if parent is not None and not hasattr(node, 'origin_time'):
                if None not in (
                    getattr(parent, 'origin_time', None),
                    node.edge_length,
                ) and not isanan(parent.origin_time):
                    node.origin_time = parent.origin_time + node.edge_length

    # attach ids to nodes, if needed
    for fallback_id, node in enumerate(tree):
        if not hasattr(node, 'id'):
            node.id = fallback_id
        else:
            assert isinstance(node.id, int)

    assert not any((
        attr in (
            exportattrs.values()
            if isinstance(exportattrs, Mapping)
            else opyt.or_value(exportattrs, [])
        )
        for attr in ('origin_time', 'id', 'edge_length', 'label', 'taxon.label')
    ))
    return pd.DataFrame.from_records([
        {
            **{
                'id': node.id,
                'ancestor_list': str(
                    [opyt.apply_if(node.parent_node, lambda x: x.id)]
                ),
                'origin_time': getattr(node, 'origin_time', None),
                'edge_length': node.edge_length,
                'label': node.label,
                'taxon_label': opyt.apply_if(node.taxon, lambda x: x.label),
            },
            **{
                exportattrs[attr_name]
                if isinstance(exportattrs, Mapping)
                else attr_name: _rgetattr(node, attr_name)
                for attr_name in opyt.or_value(exportattrs, [])
            },
        }
        for node in tree
    ])



from Bio import Phylo
from iterpop import iterpop as ip
import pandas as pd
import typing

from .alife_dataframe_to_biopython_trees \
    import alife_dataframe_to_biopython_trees


def alife_dataframe_to_biopython_tree(
    df: pd.DataFrame,
    setattrs: typing.Optional[typing.Union[
        typing.Iterable[str],
        typing.Mapping[str, str],
    ]] = None,
    setup_branch_lengths: bool = False,
) -> typing.Optional[Phylo.BaseTree.Tree]:
    """Open a phylogeny dataframe formatted to the artificial life community
    data format standards as a biopython tree.

    Returns None if df is empty.

    If two or more clades exist that do not share a common ancestor, an
    exception will be raised.

    See Also
    ----------
    alife_dataframe_to_biopython_trees
    """
    return ip.poursingleton(
        alife_dataframe_to_biopython_trees(
            df,
            setattrs=setattrs,
            setup_branch_lengths=setup_branch_lengths,
        ),
    )



from copy import deepcopy
import dendropy
from iterpop import iterpop as ip
import itertools as it
import numpy as np
from sortedcontainers import SortedSet


def dendropy_tree_to_scipy_linkage_matrix(tree: dendropy.Tree) -> np.array:
    # scipy linkage format
    # http://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html#scipy.cluster.hierarchy.linkage

    # simplify tree
    tree = deepcopy(tree)
    tree.resolve_polytomies()
    tree.suppress_unifurcations()
    assert all(len(node.child_nodes()) in [0, 2] for node in tree)

    for node in tree.postorder_node_iter():
        node.num_leaf_descendants = max(
            sum(chld.num_leaf_descendants for chld in node.child_node_iter()),
            1,
        )

    cluster_id_generator = it.count()
    for leaf in tree.leaf_node_iter():
        if not hasattr(leaf, 'cluster_id'):
            leaf.cluster_id = next(cluster_id_generator)

    one_leaf_parents = {
        leaf.parent_node
        for leaf in tree.leaf_node_iter()
        if leaf.parent_node is not None
        and not ip.popsingleton(leaf.sibling_nodes()).is_leaf()
    }
    two_leaf_parents = SortedSet(
        (
            leaf.parent_node
            for leaf in tree.leaf_node_iter()
            if leaf.parent_node is not None
            and ip.popsingleton(leaf.sibling_nodes()).is_leaf()
        ),
        key=lambda node: sum(n.edge_length for n in node.child_node_iter()),
    )

    res = []
    while len(two_leaf_parents):
        two_leaf_parent = two_leaf_parents.pop(0)
        if not hasattr(two_leaf_parent, 'cluster_id'):
            two_leaf_parent.cluster_id = next(cluster_id_generator)
        if two_leaf_parent.parent_node is not None:
            if two_leaf_parent.parent_node in one_leaf_parents:
                one_leaf_parents.remove(two_leaf_parent.parent_node)
                two_leaf_parents.add(two_leaf_parent.parent_node)
            else:
                one_leaf_parents.add(two_leaf_parent.parent_node)

        child1, child2 = two_leaf_parent.child_node_iter()
        assert child1 not in two_leaf_parents
        assert child2 not in two_leaf_parents

        # see https://stackoverflow.com/a/40983611/17332200
        # for explainer on scipy linkage format
        joined_cluster1 = child1.cluster_id
        joined_cluster2 = child2.cluster_id
        assert None not in (joined_cluster1, joined_cluster2)
        cluster_distance = child1.edge_length + child2.edge_length
        cluster_size = two_leaf_parent.num_leaf_descendants

        res.append([
            joined_cluster1,
            joined_cluster2,
            float(cluster_distance),
            cluster_size,
        ])

    return np.array(res)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "biopython>=1.79",
    "click>=7.0",
    "dendropy>=4.5.2",
    "iterpop>=0.4.0",
    "lyncs-utils>=0.3.2",
    "nanto>=0.1.1",
    "numpy>=1.21.5",
    "opytional>=0.1.0",
    "pandas>=1.1.0",
    "sortedcontainers>=2.4.0",
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Matthew Andres Moreno",
    author_email='m.more500@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="alifedata-phyloinformatics-convert helps apply traditional phyloinformatics software to alife standardized data",
    entry_points='''
    [console_scripts]
    alifedata-phyloinformatics-convert=alifedata_phyloinformatics_convert.cli:cli
    ''',
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='alifedata-phyloinformatics-convert',
    name='alifedata-phyloinformatics-convert',
    packages=find_packages(include=['alifedata_phyloinformatics_convert', 'alifedata_phyloinformatics_convert.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/mmore500/alifedata-phyloinformatics-convert',
    version='0.7.1',
    zip_safe=False,
)
